package comp3804a3;

import java.util.AbstractQueue;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Random;

/**
 * This class implements a priority queue as a class binary heap stored
 * implicitly in an array
 *
 * @author morin
 * @author Sander Verdonschot
 *
 * @param <T>
 */
public class BinaryHeap<T> extends AbstractQueue<T> {

    /**
     * Our backing array
     */
    private QueueObject<T>[] elements;

    /**
     * The number of elements in the priority queue
     */
    private int n;

    /**
     * A map connecting objects stored in this priority queue to their
     * containers in the backing array
     */
    private final Map<T, QueueObject<T>> elementMap;

    /**
     * Create a new empty binary heap.
     */
    @SuppressWarnings("unchecked")
    public BinaryHeap() {
        elements = new QueueObject[1];
        n = 0;
        elementMap = new HashMap<>();
    }

    @Override
    @SuppressWarnings("unchecked")
    public void clear() {
        elements = new QueueObject[1];
        n = 0;
        elementMap.clear();
    }

    @Override
    public int size() {
        return n;
    }

    @Override
    public boolean offer(T x) {
        return add(x);
    }

    @Override
    public boolean add(T x) {
        return add(x, Double.POSITIVE_INFINITY);
    }

    /**
     * Add element x with the given priority.
     *
     * @param x
     * @param priority
     * @return
     */
    public boolean add(T x, double priority) {
        if (x == null) {
            throw new NullPointerException();
        }

        if (n + 1 > elements.length) {
            resize();
        }

        QueueObject<T> qo = new QueueObject<>(x, priority, n);
        elementMap.put(x, qo);
        elements[n] = qo;
        n++;

        bubbleUp(n - 1);

        return true;
    }

    /**
     * Change the priority of x.
     *
     * @param x
     * @param newPriority
     */
    public void setPriority(T x, double newPriority) {
        QueueObject<T> qo = elementMap.get(x);

        boolean increased = newPriority > qo.priority;
        qo.priority = newPriority;

        if (increased) {
            trickleDown(qo.index);
        } else {
            bubbleUp(qo.index);
        }
    }

    @Override
    public T peek() {
        return elements[0].element;
    }

    @Override
    public T poll() {
        return remove();
    }

    @Override
    public T remove() {
        T x = elements[0].element;
        elementMap.remove(x);

        elements[0] = elements[n - 1];
        elements[0].index = 0;
        n--;

        trickleDown(0);

        if (3 * n < elements.length) {
            resize();
        }

        return x;
    }

    @Override
    public Iterator<T> iterator() {
        return (Iterator<T>) new PQI();
    }

    private void resize() {
        elements = Arrays.copyOf(elements, Math.max(2 * n, 1));
    }

    /**
     * @param i
     * @return the index of the left child of the value at index i
     */
    private int left(int i) {
        return 2 * i + 1;
    }

    /**
     * @param i
     * @return the index of the left child of the value at index i
     */
    private int right(int i) {
        return 2 * i + 2;
    }

    /**
     * @param i
     * @return the index of the parent of the value at index i
     */
    private int parent(int i) {
        return (i - 1) / 2;
    }

    /**
     * Swap the two values a[i] and a[j]
     *
     * @param i
     * @param j
     */
    private void swap(int i, int j) {
        QueueObject<T> x = elements[i];
        elements[i] = elements[j];
        elements[j] = x;

        elements[i].index = i;
        elements[j].index = j;
    }

    /**
     * Compare the priorities of a[i] and a[j].
     *
     * @param i
     * @param j
     * @return
     */
    private int compare(int i, int j) {
        return Double.compare(elements[i].priority, elements[j].priority);
    }

    /**
     * Run the bubbleUp routine at position i
     *
     * @param i
     */
    private void bubbleUp(int i) {
        int p = parent(i);

        while (i > 0 && compare(i, p) < 0) {
            swap(i, p);

            i = p;
            p = parent(i);
        }
    }

    /**
     * Move element i down in the heap until the heap property is restored
     *
     * @param i
     */
    private void trickleDown(int i) {
        do {
            int j = -1;
            int r = right(i);

            if (r < n && compare(r, i) < 0) {
                int l = left(i);

                if (compare(l, r) < 0) {
                    j = l;
                } else {
                    j = r;
                }
            } else {
                int l = left(i);

                if (l < n && compare(l, i) < 0) {
                    j = l;
                }
            }

            if (j >= 0) {
                swap(i, j);
            }

            i = j;
        } while (i >= 0);
    }

    private class QueueObject<T> {

        T element;
        double priority;
        int index;

        private QueueObject(T element, double priority, int index) {
            this.element = element;
            this.priority = priority;
            this.index = index;
        }
    }

    private class PQI implements Iterator<T> {

        int i;

        private PQI() {
            i = 0;
        }

        @Override
        public boolean hasNext() {
            return i < n;
        }

        @Override
        public T next() {
            i++;
            return elements[i - 1].element;
        }

        @Override
        public void remove() {
            throw new UnsupportedOperationException();
        }
    }

}
